import ipywidgets as widgets
from toolz import first
from tkinter import Tk
from tkinter.filedialog import askopenfilename
from functools import partial

"""Avoided Combobox widget because it was not rendering correctly in the notebook
    options -> value populated in Select Multi, Combo, and Dropdown
    value(s)-> Current Selected List, item, or text
    disabled -> if Selectable when rendered
    description puts label to the left of the widget.  I elected to not include
    placeholder -> value displayed when empty
"""


class WidgetLayouts:
    """Layout is used to set specific CSS settings for almost any widget
    This function sets default css settings for all shapes
    """

    @staticmethod
    def make_border_layout(width="max-contnet", margin="5px 5px 5px 5px"):
        """Layout for Model Portfolio Box and For Model Portfolio File Browser Label"""
        return {
            "border": "1px solid grey",
            "height": "max-content",
            "width": width,
            "margin": margin,
            "padding": "0px 0px 0px 0px",
        }

    def make_layout(
        height="100%",
        width="max-content",
        max_height="",
        max_width="",
        min_height="",
        min_width="",
        padding="0px 0px 0px 0px",
        margin="0px 0px 0px 0px",
    ):
        return {
            "height": height,
            "width": width,
            "max_height": max_height,
            "max_width": max_width,
            "min_height": min_height,
            "min_width": min_width,
            "padding": padding,
            "margin": margin,
        }

    @staticmethod
    def make_tab_box_layout():
        """Used for all tab boxes"""
        return WidgetLayouts.make_layout(min_width="1250px", height="max-content")


class WidgetContainers:
    @staticmethod
    def make_hbox(children=[], layout=WidgetLayouts.make_layout()):
        """horizontally stacked layout box"""
        return widgets.HBox(children=children, layout=layout)

    @staticmethod
    def make_tab(children, titles, layout=WidgetLayouts.make_layout()):
        """Tab Widget used to seperate grouped inputs"""
        tab = widgets.Tab(children=children, layout=layout)
        for idx, val in enumerate(titles):
            tab.set_title(idx, val)
        return tab

    @staticmethod
    def make_vbox(children=[], layout=WidgetLayouts.make_layout()):
        """vertically stacked layout box"""
        return widgets.VBox(children=children, layout=layout)

    @staticmethod
    def make_vbox_container(children):
        """Used to set vertical layout for each Tab Container"""
        return WidgetContainers.make_vbox(
            children=children,
            layout=WidgetLayouts.make_layout(height="475px", width="100%"),
        )


class BasicWidgets:
    default_label_font_color = "#4B4B4B"
    default_label_font_family = "Arial"

    @staticmethod
    def get_output():
        """Used to Allow Display to Interact with Widgets"""
        return widgets.Output()

    @staticmethod
    def ask_for_file(file_types=[("CSV file", "*.txt"),("Text file", "*.txt"), ]):
        """Opens a dialog to allow the user to browse for a file.  Returns a tuple boolean and filename"""
        root = Tk()
        root.attributes("-topmost", True)
        root.withdraw()
        filename = askopenfilename(filetypes=file_types, parent=root)

        if filename:
            return True, filename
        else:
            return False, ""

    @staticmethod
    def make_button(
        description,
        layout=WidgetLayouts.make_layout(
            height="30px",
            width="max-content",
            margin="5px 10px 0px 0px",
            padding="0 2px 0 2px",
        ),
        disabled=False,
    ):
        """Button Widget"""
        return widgets.Button(
            description=description,
            disabled=disabled,
            layout=layout,
        )

    # make checkbox
    def make_checkbox(description, layout=WidgetLayouts.make_layout(height="30px")):
        """Checkbox Widget"""
        return widgets.Checkbox(
            value=False, description=description, disabled=False, indent=False
        )

    @staticmethod
    def make_dropdown(
        options,
        value="",
        disabled=False,
        layout=WidgetLayouts.make_layout(height="30px", min_width="200px"),
    ):
        """Dropdown widget"""
        """Used In Place of Combobox because of rendering issues"""
        return widgets.Dropdown(
            options=options, value=value, disabled=disabled, layout=layout
        )

    @staticmethod
    def make_html_table(
        df=None, layout=WidgetLayouts.make_layout(min_width="200px", height="400px")
    ):
        """Create Formatted HTML Table from dataframe"""
        if not df is None:
            return widgets.HTML(value=df.to_html())
        else:
            return widgets.HTML()

    # Labels
    @staticmethod
    def make_alert_label(
        alert_class_prefix,
        alert_description,
        description_color,
        alert_value,
        value_color="#4B4B4B",
    ):
        """Used for Creating Various Label Types"""
        return widgets.HTML(
            value=f"""<style>
                        span{{
                            font-family:{BasicWidgets.default_label_font_family};
                            font-size:13;
                            font-weight:normal;                            
                            word-wrap:break-word;
                            width:100%;
                            line-height:normal}}
                        span.{alert_class_prefix}_description{{
                            font-weight:bold;
                            color:{description_color};
                        }}
                        span.{alert_class_prefix}_value{{
                            color:{value_color};
                        }}
                    </style>
                    <p><span class={alert_class_prefix}_description>{alert_description}</span>
                        <span class={alert_class_prefix}_value>{alert_value}</span>
                    </p>"""
        )

    @staticmethod
    def make_description_label(
        text, padding="0px 0px 0px 0px", label_class="description", width="900px"
    ):
        """Used to standardize description format (size and font weight)
        For Bolded Font inside of each Tab Box
        """
        return BasicWidgets.make_label(
            text,
            width=width,
            label_class=label_class,
            font_weight="normal",
            font_size="15px",
            line_height="normal",
            padding=padding,
        )

    @staticmethod
    def make_display_label(
        value="", layout=WidgetLayouts.make_layout(min_width="200px")
    ):
        """Used to Display Values for File Browser Widgets"""
        return widgets.Label(value=value, layout=layout)

    @staticmethod
    def make_error_label(alert_description, alert_value):
        """Return Red Heading (alert_Description) with Detail text (alert_value)"""
        return BasicWidgets.make_alert_label(
            alert_class_prefix="error",
            alert_description=alert_description,
            description_color="#FF0000",
            alert_value=alert_value,
        )

    @staticmethod
    def make_header_label(text, padding="0px 0px 0px 0px"):
        """Used For Intermediate Steps within Request Headers"""
        return BasicWidgets.make_label(
            label_class="headers", text=text, font_weight="bold", padding=padding
        )

    @staticmethod
    def make_label(
        text,
        label_class="label",
        width="",
        font_weight="normal",
        color=default_label_font_color,
        font_family=default_label_font_family,
        font_size="13px",
        word_wrap="break-word",
        line_height="1%",
        padding="0px 0px 0px 0px",
    ):
        """In order to use rich text I had to use HTML instead of standard label widget"""

        return widgets.HTML(
            value=f"""<style>
                        p.{label_class}{{
                            font-family:{font_family};
                            font-size:{font_size};
                            font-weight:{font_weight};
                            color:{color};
                            word-wrap:{word_wrap};
                            width:{width};  
                            line-height:{line_height};
                            padding:{padding};                     
                            }}
                    </style>
                    <p class={label_class}>{text}</p>"""
        )

    @staticmethod
    def make_request_label(text):
        """Blue Label Used To Indicate a Request has Started"""
        return BasicWidgets.make_label(
            label_class="request",
            text=text,
            font_weight="bold",
            color="#00447c",
            font_size="15px",
            padding="10px 0px 10px 0px",
        )

    @staticmethod
    def make_success_request_label(text):
        """Make a Green Label"""
        return BasicWidgets.make_label(
            label_class="suc_request",
            text=text,
            font_weight="bold",
            color="#008000",
            font_size="13px",
        )

    @staticmethod
    def make_selectmultiple_widget(
        options=[],
        values=[],
        layout=WidgetLayouts.make_layout(height="100px"),
        disabled=False,
    ):
        """Select Multiple Used to make and display selections"""
        return widgets.SelectMultiple(
            options=options, disabled=disabled, layout=layout, values=values
        )

    @staticmethod
    def make_textbox(
        placeholder="", value="", layout=WidgetLayouts.make_layout(min_width="200px")
    ):
        """text box widget"""
        return widgets.Text(value=value, placeholder=placeholder, layout=layout)


class WidgetCombinations:
    @staticmethod
    def make_category_item_box(
        components_dictionary,
        category_label="",
        item_label="",
        selected_label="",
        single_select=False,
        uncategorized_additional_dictionary=None,
        additional_info_label_text="",
        text_input_label="",
        dict_info=None,
    ):
        """
        This fucntion is used to create mutliple visualizations:
        At its simpliest form (ommiting the uncategorized dictionary and dict_ifo)
        It creates a Category Drop Down, Item Dropdown, and a selectmultiple (used to display the selected values)
        Setting single select to true will only allow the user to select a single value

        Adding an uncategorized additional dictionary will create a second SelectMulti where the user can assign additional attributes for the selected value (Column Statistics)

        Adding text_input_label / dict_info will create a text input box and get button.  This allows the user to specify the source or the category and item dropdowns
            when included the response will include
                category_item_box,
                selectmultiple,
                category_dropdown,
                item_dropdown,
                input_text,
                retrieve_button,
            When excluded:

            This will allow you to specify the events for the retrieve button and set the values for the underlying components
            This dict_info will allow the user to maintain the results of previous requests locally to avoid having to repeatedly call the downstream process to populate the form.

        The SelectMultiple will store all information for the selected item including it's category, additional information, and "Text Input" (when required) in a tuple.
        the tuple should be (Value Displayed,"Category", Description, ID (results), and when required the text field value

        )
        """
        additional_info_selectmultiple = None
        additional_info_box = None
        input_box = None
        input_text = None
        if text_input_label:
            (
                input_box,
                input_text,
                retrieve_button,
            ) = WidgetCombinations.make_text_lookup_box("PA Document", "GET")

        if uncategorized_additional_dictionary:
            (
                additional_info_box,
                additional_info_selectmultiple,
            ) = WidgetCombinations.make_multiselect_selectmultiple(
                label=additional_info_label_text,
                options=list(uncategorized_additional_dictionary),
                layout=WidgetLayouts.make_layout(
                    height="100%",
                    min_width="650px",
                    width="max-content",
                    margin="0px 10px 0px 0px",
                ),
            )

        category_label = BasicWidgets.make_header_label(
            category_label
        )  # BasicWidgets.make_label(category_label)
        category_dropdown = BasicWidgets.make_dropdown(
            options=components_dictionary, value=[("", "")]
        )
        item_label = BasicWidgets.make_header_label(
            item_label
        )  # BasicWidgets.make_label(item_label)
        item_dropdown = BasicWidgets.make_dropdown(options=[""])
        (
            selectmultiple_box,
            selectmultiple,
        ) = WidgetCombinations.make_display_selectmultiple(
            layout=WidgetLayouts.make_layout(
                height="100%",
                min_width="650px",
                width="max-content",
                margin="0px 10px 0px 0px",
            )
            if not single_select
            else WidgetLayouts.make_layout(
                height="30px",
                min_width="650px",
                width="max-content",
                margin="0px 10px 0px 0px",
            ),
            label=selected_label,
            item_dropdown=item_dropdown,
        )

        if input_box:
            category_item_box = WidgetContainers.make_vbox(
                children=[
                    input_box,
                    category_label,
                    category_dropdown,
                    item_label,
                    item_dropdown,
                    WidgetContainers.make_hbox(
                        children=[
                            selectmultiple_box,
                            additional_info_box,
                        ],
                        layout=WidgetLayouts.make_layout(),
                    )
                    if uncategorized_additional_dictionary
                    else selectmultiple_box,
                ]
            )
        else:
            category_item_box = WidgetContainers.make_vbox(
                children=[
                    category_label,
                    category_dropdown,
                    item_label,
                    item_dropdown,
                    WidgetContainers.make_hbox(
                        children=[
                            selectmultiple_box,
                            additional_info_box,
                        ]
                    )
                    if uncategorized_additional_dictionary
                    else selectmultiple_box,
                ]
            )

        item_callback = partial(
            WidgetEvents.add_value_to_selectmultiple_drop_down,
            display_selectmultiple=selectmultiple,
            single_select=single_select,
            is_tuple=[True if uncategorized_additional_dictionary else False],
        )
        selectmutliple_callback = partial(
            WidgetEvents.display_selectmultiple_changed,
            category_dropdown=category_dropdown,
            item_dropdown=item_dropdown,
            selectmultiple_info=additional_info_selectmultiple,
            item_callback=item_callback,
            input_text=input_text,
            dict_info=dict_info,
        )

        selectmultiple.observe(selectmutliple_callback)

        category_dropdown.observe(
            partial(
                WidgetEvents.category_dropdown_changed,
                component_dropdown=item_dropdown,
                component_callback=item_callback,
                selectmultiple=selectmultiple,
                selectmultiple_info=additional_info_selectmultiple,
            )
        )
        item_dropdown.observe(item_callback)
        if uncategorized_additional_dictionary:
            additional_info_selectmultiple.observe(
                partial(
                    WidgetEvents.additional_info_selected,
                    selectmultiple_item=selectmultiple,
                    category_dropdown=category_dropdown,
                    item_dropdown=item_dropdown,
                    item_callback=item_callback,
                )
            )
        if input_box:
            return (
                category_item_box,
                selectmultiple,
                category_dropdown,
                item_dropdown,
                input_text,
                retrieve_button,
            )
        else:
            return category_item_box, selectmultiple

    @staticmethod
    def make_directory_category_item_box(
        components_dictionary,
        description="",
        directory_label="",
        category_label="",
        item_label="",
        selected_label="",
        single_select=False,
        uncategorized_additional_dictionary=None,
        dict_info=None,
    ):
        """
        This fucntion is used to create mutliple visualizations:
        At its simpliest form (ommiting the uncategorized dictionary and dict_ifo)
        It creates a Category Drop Down, Item Dropdown, and a selectmultiple (used to display the selected values)
        Setting single select to true will only allow the user to select a single value

        Adding an uncategorized additional dictionary will create a second SelectMulti where the user can assign additional attributes for the selected value (Column Statistics)

        Adding text_input_label / dict_info will create a text input box and get button.  This allows the user to specify the source or the category and item dropdowns
            when included the response will include
                category_item_box,
                selectmultiple,
                category_dropdown,
                item_dropdown,
                input_text,
                retrieve_button,
            When excluded:

            This will allow you to specify the events for the retrieve button and set the values for the underlying components
            This dict_info will allow the user to maintain the results of previous requests locally to avoid having to repeatedly call the downstream process to populate the form.

        The SelectMultiple will store all information for the selected item including it's category, additional information, and "Text Input" (when required) in a tuple.
        the tuple should be (Value Displayed,"Category", Description, ID (results), and when required the text field value

        )
        """
        description_label = BasicWidgets.make_header_label(description)

        directory_label = BasicWidgets.make_header_label(directory_label)
        directory_dropdown = BasicWidgets.make_dropdown(
            options=components_dictionary, value={"": [("", "")]}
        )

        options = {"": [("", "")]}
        category_label = BasicWidgets.make_header_label(category_label)
        category_dropdown = BasicWidgets.make_dropdown(
            options=options, value=[("", "")]
        )
        item_label = BasicWidgets.make_header_label(item_label)
        item_dropdown = BasicWidgets.make_dropdown(options=[("", "")])
        (
            selectmultiple_box,
            selectmultiple,
        ) = WidgetCombinations.make_display_selectmultiple(
            layout=WidgetLayouts.make_layout(
                height="100%", min_width="500px", width="max-content"
            )
            if not single_select
            else WidgetLayouts.make_layout(
                height="30px", min_width="500px", width="max-content"
            ),
            label=selected_label,
            item_dropdown=item_dropdown,
        )

        category_item_box = WidgetContainers.make_vbox(
            children=[
                description_label,
                directory_label,
                directory_dropdown,
                category_label,
                category_dropdown,
                item_label,
                item_dropdown,
                selectmultiple_box,
            ],
            layout=WidgetLayouts.make_layout(
                padding="0px 10px 0px 0px", min_width="500px"
            ),
        )

        item_callback = partial(
            WidgetEvents.add_value_to_selectmultiple_drop_down,
            display_selectmultiple=selectmultiple,
            single_select=single_select,
            is_tuple=[True if uncategorized_additional_dictionary else False],
        )
        # 'Need to Update'

        category_callback = partial(
            WidgetEvents.category_dropdown_changed,
            component_dropdown=item_dropdown,
            component_callback=item_callback,
            selectmultiple=selectmultiple,
        )

        directory_callback = partial(
            WidgetEvents.directory_dropdown_changed,
            category_dropdown=category_dropdown,
            category_callback=category_callback,
            component_dropdown=item_dropdown,
            component_callback=item_callback,
            selectmultiple=selectmultiple,
        )

        category_dropdown.observe(category_callback)

        item_dropdown.observe(item_callback)

        directory_dropdown.observe(directory_callback)

        selectmutliple_callback = partial(
            WidgetEvents.directory_display_selectmultiple_changed,
            directory_dropdown=directory_dropdown,
            directory_callback=directory_callback,
            category_dropdown=category_dropdown,
            category_callback=category_callback,
            item_dropdown=item_dropdown,
            item_callback=item_callback,
        )

        selectmultiple.observe(selectmutliple_callback)

        return category_item_box, selectmultiple

    @staticmethod
    def make_display_selectmultiple(
        layout=WidgetLayouts.make_layout(width="500px"),
        label="",
        item_dropdown=None,
    ):
        """
        Vertically Stacks Label, SelectMultiple and Remove Button
        Creates Event handler for Remove Button for SelectMultiple.
        Couldn't use make_selectmultiple label because of the vertical alignment when setting Vbox inside of Vbox
        Returns vBox and SelectMultiple in Tuple
        """
        output_selectmultiple = BasicWidgets.make_selectmultiple_widget(layout=layout)
        remove_button = BasicWidgets.make_button(description="Remove")
        remove_button.on_click(
            partial(
                WidgetEvents.remove_display_selectmultiple,
                selectmultiple_values=output_selectmultiple,
                item_dropdown=item_dropdown,
            )
        )
        box_vertical = WidgetContainers.make_vbox(
            children=[
                BasicWidgets.make_header_label(label),
                output_selectmultiple,
                remove_button,
            ]
        )
        return box_vertical, output_selectmultiple

    @staticmethod
    def make_dropdown_with_label(
        label="", options=[], value="", layout=WidgetLayouts.make_layout(height="30px")
    ):
        """
        Stacks Label ontop of Dropdown
        Returns vBox Container and DropDown in Tuple
        """
        dd_label = BasicWidgets.make_header_label(label)
        dropdown = BasicWidgets.make_dropdown(
            options=options, value=value, layout=layout
        )
        box = WidgetContainers.make_vbox(children=[dd_label, dropdown])
        return box, dropdown

    @staticmethod
    def make_label_lookup_box(
        box_label,
        button_label,
        label_input_box_layout=WidgetLayouts.make_layout(
            height="30px", width="max-content"
        ),
        default_value="",
    ):
        """Used for File Browser UI"""
        input_label = BasicWidgets.make_header_label(box_label)
        output_label = BasicWidgets.make_display_label(
            value=default_value,
            layout=label_input_box_layout,
        )
        output_box = WidgetContainers.make_vbox(
            children=[output_label],
            layout=WidgetLayouts.make_border_layout(margin="0px 10px 0px 0px"),
        )

        retrieve_button = BasicWidgets.make_button(
            description=button_label,
            layout=WidgetLayouts.make_layout(height="30px"),
        )
        input_box = WidgetContainers.make_vbox(
            [
                input_label,
                WidgetContainers.make_hbox(
                    children=[output_box, retrieve_button],
                    layout=WidgetLayouts.make_layout(),
                ),
            ],
            layout=WidgetLayouts.make_layout(padding="0px 0px 0px 10px"),
        )
        return input_box, output_label, retrieve_button

    @staticmethod
    def make_multiselect_selectmultiple(
        layout=WidgetLayouts.make_layout(width="400px"), options=[], label=""
    ):
        """Used for Column Statistcs"""
        output_selectmultiple = BasicWidgets.make_selectmultiple_widget(
            layout=layout, options=options
        )
        remove_button = BasicWidgets.make_button(description="Clear All")

        remove_button.on_click(
            partial(
                WidgetEvents.clear_values_selectmultiple,
                selectmultiple_widget=output_selectmultiple,
            )
        )
        box_vertical = WidgetContainers.make_vbox(
            children=[
                BasicWidgets.make_header_label(text=label),
                output_selectmultiple,
                remove_button,
            ]
        )
        return box_vertical, output_selectmultiple

    @staticmethod
    def make_selectmultiple_with_label(
        label, options=[], values=[], disabled=False, layout=WidgetLayouts.make_layout()
    ):
        """
        Vertically Stacks label ontop of SelectMultiple
        Returns vBox container and SelectMultiple in Tuple
        """
        selectmultiple_label = BasicWidgets.make_label(label)
        selectmultiple = BasicWidgets.make_selectmultiple_widget(
            options=options, values=values, layout=layout
        )
        return (
            WidgetContainers.make_vbox(children=[selectmultiple_label, selectmultiple]),
            selectmultiple,
        )

    @staticmethod
    def make_textbox_with_label(
        label,
        placeholder="",
        value="",
        layout=WidgetLayouts.make_layout(),
        text_box_width="240px",
    ):
        """
        Vertically stacks a label ontop of textbox
        Returns the vBox Container and the TextBox in tuple
        """
        text_label = BasicWidgets.make_header_label(
            text=label, padding="0px 0px 0px 0px"
        )
        text_box = BasicWidgets.make_textbox(
            placeholder=placeholder,
            value=value,
            layout=WidgetLayouts.make_layout(width=text_box_width),
        )

        return (
            WidgetContainers.make_vbox(children=[text_label, text_box], layout=layout),
            text_box,
        )

    @staticmethod
    def make_text_lookup_box(
        box_label,
        button_label,
        text_input_box_layout=WidgetLayouts.make_layout(
            height="40px", width="400px", margin="0px 10px 0px 0px"
        ),
        default_value="",
    ):
        """Originally Used for File Browser.  Just kept in case needed"""

        input_label = BasicWidgets.make_label(box_label)
        input_text = BasicWidgets.make_textbox(
            value=default_value,
            placeholder="",
            layout=text_input_box_layout,
        )
        retrieve_button = BasicWidgets.make_button(
            description=button_label,
            layout=WidgetLayouts.make_layout(height="30px"),
        )
        input_box = WidgetContainers.make_vbox(
            [
                input_label,
                WidgetContainers.make_hbox(
                    children=[input_text, retrieve_button],
                    layout=WidgetLayouts.make_layout(),
                ),
            ]
        )
        return input_box, input_text, retrieve_button

    @staticmethod
    def make_text_input_box(
        text_label="",
        text_value="",
        placeholder="",
        selectmultiple_label="",
        layout=WidgetLayouts.make_layout(),
    ):
        """
        Creates a stacked Label, text box with add button and select multiple with remove button
        """

        input_label = BasicWidgets.make_label(text_label)
        input_text = BasicWidgets.make_textbox(
            value=text_value,
            placeholder=placeholder,
            layout=WidgetLayouts.make_layout(
                height="30px", width="300px", margin="0px 5px 0px 0px"
            ),
        )
        add_button = BasicWidgets.make_button(
            description="ADD",
            layout=WidgetLayouts.make_layout(height="30px", width="30px"),
        )
        input_box = WidgetContainers.make_hbox(children=[input_text, add_button])
        (
            selectmultiple_box,
            selectmultiple,
        ) = WidgetCombinations.make_display_selectmultiple(
            layout=WidgetLayouts.make_layout(height="75px", width="300px"),
            label=selectmultiple_label,
        )
        box = WidgetContainers.make_vbox(
            children=[input_label, input_box, selectmultiple_box],
            layout=layout,
        )
        add_button.on_click(
            partial(
                WidgetEvents.add_value_to_selectmulitple_text,
                text_box=input_text,
                display_selectmultiple=selectmultiple,
            )
        )
        selectmultiple.observe(
            partial(
                WidgetEvents.display_selectmultiple__text_changed,
                input_textbox=input_text,
            )
        )
        return box, selectmultiple


class WidgetEvents:
    @staticmethod
    def remove_display_selectmultiple(b, selectmultiple_values, item_dropdown=None):
        """Event which removes selected values from selectmultiple"""
        selected_values = selectmultiple_values.value
        if selected_values:
            selectmultiple_values.options = filter(
                lambda i: i[1] not in selected_values, selectmultiple_values.options
            )
            if not item_dropdown is None:
                item_dropdown.value = ""

    @staticmethod
    def add_value_to_selectmulitple_text(
        b, text_box, display_selectmultiple, single_select=False
    ):
        """Add Value from TextBox to SelectMultiple"""
        text_value = text_box.value.strip()
        if len(text_value) > 0:
            text_box_value = (text_value, text_value)
            if text_box_value not in list(display_selectmultiple.options):
                display_selectmultiple.options = (
                    [text_box_value]
                    if single_select
                    else list(display_selectmultiple.options) + [text_box_value]
                )
                text_box.value = ""

    @staticmethod
    def category_dropdown_changed(
        change,
        component_dropdown,
        component_callback,
        selectmultiple,
        selectmultiple_info="",
    ):
        """
        Update the component_dropdown with values from dropdown who triggered the event
        the Value stored in the event dropdown must be the list of values ues to populate the component_dropdown
        must unobserve during update to prevent incorrect selections
        """
        if change["type"] == "change" and change["name"] == "value":
            list_values = change["new"]
            component_dropdown.unobserve(component_callback)
            component_dropdown.options = sorted(list_values)
            component_dropdown.observe(component_callback)
            selectmultiple.value = []
            if selectmultiple_info:
                selectmultiple_info.value = []

    def directory_dropdown_changed(
        change,
        category_dropdown,
        category_callback,
        component_dropdown,
        component_callback,
        selectmultiple,
        selectmultiple_info="",
    ):
        """
        Update the component_dropdown with values from dropdown who triggered the event
        the Value stored in the event dropdown must be the list of values ues to populate the component_dropdown
        must unobserve during update to prevent incorrect selections
        """
        if change["type"] == "change" and change["name"] == "value":
            list_values = change["new"]
            category_dropdown.unobserve(category_callback)
            category_dropdown.options = sorted(list_values.items())
            category_dropdown.observe(category_callback)
            component_dropdown.unobserve(component_callback)
            component_dropdown.options = [("", "")]
            component_dropdown.observe(component_callback)
            selectmultiple.value = []
            if selectmultiple_info:
                selectmultiple_info.value = []

    @staticmethod
    def clear_values_selectmultiple(b, selectmultiple_widget):
        selectmultiple_widget.value = []

    # Add value from dropdown to Display Select multiple
    @staticmethod
    def add_value_to_selectmultiple_drop_down(
        change, display_selectmultiple, single_select=False, is_tuple=False
    ):
        if change["type"] == "change" and change["name"] == "value":
            dropdown_value = change["new"]
            if dropdown_value:
                if (
                    is_tuple
                    and not [
                        item
                        for item in list(display_selectmultiple.options)
                        if item[1][2] == dropdown_value[1][2]
                    ]
                ) or (
                    not is_tuple
                    and dropdown_value not in list(display_selectmultiple.options)
                ):
                    display_selectmultiple.options = (
                        [dropdown_value]
                        if single_select
                        else list(display_selectmultiple.options) + [dropdown_value]
                    )

                    display_selectmultiple.value = [dropdown_value[1]]

    @staticmethod
    def display_selectmultiple__text_changed(change, input_textbox):
        """
        This event allows the selected selectmultiple value to overwrite the text inside of the text input box
        """
        if change["type"] == "change" and change["name"] == "value":
            selection = list(change["new"])
            if selection:
                if len(selection) == 1:
                    input_textbox.value = first(selection)

    @staticmethod
    def directory_display_selectmultiple_changed(
        change,
        directory_dropdown,
        directory_callback,
        category_dropdown,
        category_callback,
        item_dropdown,
        item_callback,
    ):
        """
        This event allows the display_selectmuitples to return the Item and category and change the parent dropdowns (Category and Item)
        """
        if change["type"] == "change" and change["name"] == "value":
            selection = list(change["new"])
            if selection:
                # Only update item and category when item is selected
                if len(selection) == 1:
                    # Remove event listener
                    item_dropdown.unobserve(None)
                    # Change value of category dropdown
                    category_dropdown.unobserve(None)
                    directory_dropdown.unobserve(None)

                    directory_dropdown.value = directory_dropdown.options[
                        first(selection)[0]
                    ]
                    category_dropdown.options = directory_dropdown.value
                    category_dropdown.value = category_dropdown.options[
                        first(selection)[1]
                    ]
                    item_dropdown.options = category_dropdown.value
                    item_dropdown.value = first(
                        [
                            item
                            for item in item_dropdown.options
                            if item[0] == first(selection)[2]
                        ]
                    )[1]
                    item_dropdown.observe(item_callback)
                    category_dropdown.observe(category_callback)
                    directory_dropdown.observe(directory_callback)

    @staticmethod
    def display_selectmultiple_changed(
        change,
        category_dropdown,
        item_dropdown,
        item_callback,
        selectmultiple_info=None,
        input_text=None,
        dict_info=None,
    ):
        """
        This event allows the display_selectmuitples to return the Item and category and change the parent dropdowns (Category and Item)
        This event will also change the current value in any additional information dropdowns (Columns Visual)
        (Columns Tab in Component Manager POC)
        This event assumes there are two dropdowns one for category, one for item.
        if additional info is required, has data, and only one item is selected display current value moving the results to the top of the list
        Otherwise show no selection and sort additional information alphabetically
        """

        if change["type"] == "change" and change["name"] == "value":
            selection = list(change["new"])
            if selection:
                # Only update item and category when item is selected
                if len(selection) == 1:
                    # Remove event listener
                    item_dropdown.unobserve(None)
                    # Change value of category dropdown
                    category_dropdown.unobserve(None)
                    if input_text:
                        components = dict_info[(first(selection)[3])]
                        category_dropdown.options = components
                        input_text.value = first(selection)[3]

                    category_dropdown.value = category_dropdown.options[
                        first(selection)[0]
                    ]
                    item_dropdown.options = category_dropdown.value
                    # Change Item dropdown value
                    item_dropdown.value = first(
                        [
                            item
                            for item in item_dropdown.options
                            if item[0] == first(selection)[1]
                        ]
                    )[1]
                    category_dropdown.observe(
                        partial(
                            WidgetEvents.category_dropdown_changed,
                            component_dropdown=item_dropdown,
                            component_callback=item_callback,
                            selectmultiple=change["owner"],
                            selectmultiple_info=selectmultiple_info,
                        )
                    )
                    item_dropdown.observe(item_callback)

                if selectmultiple_info:
                    if len(selection) == 1 and first(selection)[3]:
                        options = selectmultiple_info.options
                        list_selected = []
                        list_not_selected = []
                        for option in options:
                            if option[1] in first(selection)[3]:
                                list_selected.append(option)
                            elif option != ("", ""):
                                list_not_selected.append(option)

                        # Remove Event Listener for Additional Information SelectMultiple'
                        selectmultiple_info.unobserve(None)
                        # Set SelectMultiple to Reordered Values (To Display Selected On Top)'
                        selectmultiple_info.options = (
                            [("", "")] + list_selected + list_not_selected
                        )
                        # Set the SelectMultiples Values If they are currently stored in selection'
                        selectmultiple_info.value = (
                            first(selection)[3] if first(selection)[3] else []
                        )
                        # Add Event Listener Back In'
                        selectmultiple_info.observe(
                            partial(
                                WidgetEvents.additional_info_selected,
                                selectmultiple_item=change["owner"],
                                category_dropdown=category_dropdown,
                                item_dropdown=item_dropdown,
                                item_callback=item_callback,
                            )
                        )
                    else:
                        # If No Additional Info Provided
                        # Remove Event Listener
                        selectmultiple_info.unobserve(None)
                        # Reset Options to Alphabetical Order
                        selectmultiple_info.options = sorted(
                            selectmultiple_info.options
                        )
                        # Reset Values to an empty array
                        selectmultiple_info.value = []
                        # Add Back Event Listener
                        selectmultiple_info.observe(
                            partial(
                                WidgetEvents.additional_info_selected,
                                selectmultiple_item=change["owner"],
                                category_dropdown=category_dropdown,
                                item_dropdown=item_dropdown,
                                item_callback=item_callback,
                            )
                        )

    @staticmethod
    def additional_info_selected(
        change, selectmultiple_item, category_dropdown, item_dropdown, item_callback
    ):
        """Used for the Column Visual in the Component Manager POC
        This allows the user to update the Statistics Information for one more columns.
        The results are stored in the values object in selectmultiple_item
        The dropdowns are only passed to add and remove event listener
        """
        if change["type"] == "change" and change["name"] == "value":
            selected_stats = list(change["new"])
            if len(selected_stats) == 1 and first(selected_stats) == "":
                selected_stats = None

            selected_items = list(selectmultiple_item.value)
            list_selected_after_update = []
            options = list(selectmultiple_item.options)
            if selected_items:
                for index in selectmultiple_item.index:
                    options[index] = (
                        options[index][0],
                        (
                            options[index][1][0],
                            options[index][1][1],
                            options[index][1][2],
                            selected_stats,
                        ),
                    )
                    list_selected_after_update.append(
                        (
                            options[index][1][0],
                            options[index][1][1],
                            options[index][1][2],
                            selected_stats,
                        )
                    )
                if list_selected_after_update:
                    selectmultiple_item.unobserve(None)
                    selectmultiple_item.options = options
                    selectmultiple_item.value = list_selected_after_update
                    selectmultiple_item.observe(
                        partial(
                            WidgetEvents.display_selectmultiple_changed,
                            selectmultiple_info=change["owner"],
                            category_dropdown=category_dropdown,
                            item_dropdown=item_dropdown,
                            item_callback=item_callback,
                        )
                    )

    @staticmethod
    def filter_two_selelctMultiples(
        change,
        selectmulti_update,
        optional_selectmultiple,
        list_content,
    ):
        """Used when display same values in three selectmultiple widgets.  This event will ensure each of the"""
        """Used only for displaying """
        if change["type"] == "change" and change["name"] == "value":
            selection = list(change["new"])
            selectmulti_update.unobserve(None)
            selectmulti_update.value = list(
                filter(
                    lambda c: c not in selection,
                    selectmulti_update.value,
                )
            )
            optional_selectmultiple.options = list(
                filter(
                    lambda c: c not in list(selectmulti_update.value) + selection,
                    list_content,
                )
            )
            selectmulti_update.observe(
                partial(
                    WidgetEvents.filter_content,
                    selectmulti_update=change["owner"],
                    optional_selectmultiple=optional_selectmultiple,
                    list_content=list_content,
                ),
                names="value",
            )
