import re
from IPython.display import display
import json as JSON
from Common_Widgets import (
    WidgetCombinations,
    BasicWidgets,
    WidgetContainers,
    WidgetLayouts,
)
from functools import partial
from os.path import exists
import pandas as pd
from Analytics_API_Common_Functions import FDSAnalyticsAPI
from Analytics_API_Common_Widgets import AnalyticsAPIWidgets

class ModelPortfolio(AnalyticsAPIWidgets):
    __account_path_text = None
    __file_browser_label = None
    __fds_username_hyphen_serial = None
    __fds_apikey = None
    __delete_account_checkbox = None
    __display_json_results = False
    model_portfolio = None

    def __init__(
        self, fds_username_hyphen_serial, fds_apikey, display_json_results=False
    ):
        self.__fds_username_hyphen_serial = fds_username_hyphen_serial
        self.__fds_apikey = fds_apikey
        self.__display_json_results = display_json_results
        super().__init__(
            fds_username_hyphen_serial=fds_username_hyphen_serial, fds_apikey=fds_apikey
        )
        self.__renderUI()

    def __renderUI(self):
        create_submit_box = self.__make_create_submit_container()
        model_portfolio_box = self.__make_model_portfolio_box()

        outer_box = WidgetContainers.make_vbox(
            children=[
                model_portfolio_box,
                create_submit_box,
            ],
            layout=WidgetLayouts.make_border_layout(width="1200px"),
        )
        display(outer_box, self.output)

    # Model Portfolio Containers (boxes)
    def __make_create_submit_container(self):
        run_button = BasicWidgets.make_button("Create Account")
        return_json_button = BasicWidgets.make_button("Return JSON")

        self.__delete_account_checkbox = BasicWidgets.make_checkbox(
            description="Delete Account After Run"
        )
        children = [
            run_button,
            return_json_button,
            self.__delete_account_checkbox,
        ]
        button_box = WidgetContainers.make_hbox(
            children=children,
            layout=WidgetLayouts.make_layout(height="40px", padding="0px 0px 0px 10px"),
        )
        run_button.on_click(self.__run_button_on_click)
        return_json_button.on_click(self.__json_button_on_click)
        return button_box

    def __make_model_portfolio_box(self):
        df = None
        description_label = BasicWidgets.make_description_label(
            self.default_values.info_label,
            padding="0px 0px 0px 10px",
            label_class="mp_description",
        )
        # Create file Browser Text and retrieve button
        (
            file_browser_box,
            self.__file_browser_label,
            retrieve_button,
        ) = WidgetCombinations.make_label_lookup_box(
            box_label="Holdings Flat File",
            button_label="Browse",
            label_input_box_layout=WidgetLayouts.make_layout(
                height="30px", width="max-content", min_width="325px"
            ),
            default_value=self.default_values.default_filepath,
        )

        # Create Account Text input
        (
            account_name_path_box,
            self.__account_path_text,
        ) = WidgetCombinations.make_textbox_with_label(
            label="Account Name",
            layout=WidgetLayouts.make_layout(
                width="800px", padding="10px 10px 10px 10px"
            ),
            value=self.default_values.default_account_path,
        )
        # Create HTML Table from CSV
        # If Default Value Exists Load It

        if exists(self.default_values.default_filepath):
            df = FDSAnalyticsAPI.FormatResults.format_dataframe(
                pd.read_csv(self.default_values.default_filepath)
            )

        html_table = BasicWidgets.make_html_table(df=df)
        html_table_box = WidgetContainers.make_vbox(
            children=[html_table],
            layout=WidgetLayouts.make_layout(
                height="300px", margin="10px 10px 0px 10px"
            ),
        )

        retrieve_button.on_click(
            partial(
                self.__open_file_dialog,
                html_table_widget=html_table,
                file_browser_label=self.__file_browser_label,
            )
        )

        return WidgetContainers.make_vbox(
            children=[
                description_label,
                file_browser_box,
                account_name_path_box,
                html_table_box,
            ],
            layout=WidgetLayouts.make_layout(margin="0px 0px 0px 0px", width="1000px"),
        )

    # Function to Control The File Browser
    def __open_file_dialog(self, b, html_table_widget, file_browser_label):
        bfile_selected, filename = BasicWidgets.ask_for_file()
        if bfile_selected:
            file_browser_label.value = filename
            try:
                html_table_widget.value = (
                    FDSAnalyticsAPI.FormatResults.format_dataframe(
                        pd.read_csv(filename)
                    ).to_html()
                )
                self.output.clear_output()
            except:
                with self.output:
                    display(
                        BasicWidgets.make_error_label(
                            "Invalid File Format", f"The {filename} is not a valid CSV"
                        )
                    )
                    html_table_widget.value = ""

    # Dialogs for validating user inputs and displaying the PUT request information
    def __validate_return_url_json(self):
        """Validate User Inputs and Output PUT Request information"""

        if self.__account_path_text.value:
            account_path = str(self.__account_path_text.value).strip()
            # Add .ACCT extension if missing.
            self.__account_path_text.value = (
                f"{account_path}.acct"
                if account_path.upper().endswith(".ACCT") == False
                else account_path
            )
            # Create GEt URL Make sure you encode the ACCT Path
            URL = f"{FDSAnalyticsAPI.fds_api_host}/analytics/accounts/v3/models/{FDSAnalyticsAPI.Paramaters.URL_encode(self.__account_path_text.value)}"
            if self.__file_browser_label.value:
                # Check input file exists
                if exists(self.__file_browser_label.value):
                    (
                        bjson,
                        json,
                    ) = FDSAnalyticsAPI.Requests.make_model_portfolio_json_from_CSV(
                        self.__file_browser_label.value
                    )
                    if bjson:
                        with self.output:
                            display(BasicWidgets.make_header_label("PUT REQUEST URL:"))
                            print(URL)
                            display(BasicWidgets.make_header_label("PUT REQUEST BODY:"))
                            print(JSON.dumps(json, indent=2))
                            return True, URL, json
                    else:
                        with self.output:
                            display(
                                BasicWidgets.make_error_label(
                                    "Incorrect CSV Format:",
                                    json,
                                )
                            )
                        return False, "", ""

                else:
                    with self.output:
                        display(
                            BasicWidgets.make_error_label(
                                "CSV File Does Not Exist:",
                                self.__file_browser_label.value,
                            )
                        )
                        return False, "", ""
            else:
                with self.output:
                    display(
                        BasicWidgets.make_error_label(
                            "Missing CSV File:",
                            "Please Browse for or Enter the full path to the CSV file above",
                        )
                    )
                    return False, "", ""
        else:
            with self.output:
                display(
                    BasicWidgets.make_error_label(
                        "Missing Account Path:",
                        "Please enter the full path to the Account you would like to create or update above.",
                    )
                )
                return False, "", ""

    # Model Portfolio Submit Button Events
    def __json_button_on_click(self, b):
        self.output.clear_output()
        self.__validate_return_url_json()

    def __run_button_on_click(self, b):
        self.output.clear_output()
        FDSAnalyticsAPI.Requests.delete_account(fds_username_hyphen_serial=self.__fds_username_hyphen_serial, fds_apikey=self.__fds_apikey,account_name_path=self.__account_path_text.value)
        with self.output:
            display(BasicWidgets.make_request_label("Generating Model Portfolio"))
        bValid, URL, json = self.__validate_return_url_json()
        if bValid:
            (
                bSuccess,
                response_label,
                err_value,
            ) = FDSAnalyticsAPI.Requests.make_modify_model_portfolio(
                fds_username_hyphen_serial=self.__fds_username_hyphen_serial,
                fds_apikey=self.__fds_apikey,
                account_name_path=self.__account_path_text.value,
                json_string=json,
                URL=URL,
            )
            with self.output:
                if bSuccess:
                    display(BasicWidgets.make_success_request_label(response_label))
                    display(
                        BasicWidgets.make_request_label(
                            f"Requesting Model Portfolio Data: {self.__account_path_text.value}"
                        )
                    )
                    display(BasicWidgets.make_header_label("GET REQUEST URL:"))
                    (
                        bAccountReturned,
                        Account_JSON,
                        err_value,
                    ) = FDSAnalyticsAPI.LookUp.get_account_data(
                        fds_username_hyphen_serial=self.__fds_username_hyphen_serial,
                        fds_apikey=self.__fds_apikey,
                        account_name_path=self.__account_path_text.value,
                    )
                    if bAccountReturned:
                        print(err_value)
                        display(
                            BasicWidgets.make_success_request_label(
                                f"Model Portfolio {self.__account_path_text.value} Contents Successfully Retrieved (Status Code: 200)"
                            )
                        )
                        display(
                            BasicWidgets.make_header_label(
                                "Returning JSON Results as a Dataframe"
                            )
                        )
                        display(
                            FDSAnalyticsAPI.FormatResults.format_dataframe(
                                FDSAnalyticsAPI.FormatResults.return_dataframe_from_model_portfolio_response(
                                    Account_JSON
                                )
                            )
                        )
                        if self.__display_json_results:
                            display(BasicWidgets.make_header_label("JSON Results"))
                            print(JSON.dumps(Account_JSON, indent=2))

                        if self.__delete_account_checkbox.value == True:
                            self.make_pa_delete_model_portfolio_request(
                                self.__account_path_text.value
                            )
                        else:
                            self.model_portfolio = self.__account_path_text.value
                    else:
                        display(BasicWidgets.make_error_label(Account_JSON, err_value))
                else:
                    display(BasicWidgets.make_error_label(response_label, err_value))

    class default_values:
        info_label = """Browse for the text file containing the holdings you would like to upload into FactSet.  The text file must be comma delimited and contain a <strong>date</strong>, <strong>symbol</strong>, and either <strong>weights/shares</strong> columns.  Optionally, you can include a <strong>price</strong> field but our API requires a <strong>priceiso</strong> field also be included.  *Please make sure each column is labelled correctly in row one. <br><br>The Account Name (personal:sample_model_account.acct) must include a valid FactSet directory (<strong>client:/</strong>, <strong>super_client:/</strong>, or <strong>personal:/</strong>) and end with an <strong>.acct</strong> file extension."""
        default_filepath = "Sample.txt"
        default_account_path = "personal:sample_model_account.acct"
